# Copyright 2022 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import http.client

import flask
from flask import make_response
from oslo_log import log
from oslo_serialization import jsonutils

from keystone.api._shared import authentication
from keystone.api._shared import json_home_relations
from keystone.common import provider_api
from keystone.common import utils
from keystone.conf import CONF
from keystone import exception
from keystone.federation import utils as federation_utils
from keystone.i18n import _
from keystone.server import flask as ks_flask

LOG = log.getLogger(__name__)

PROVIDERS = provider_api.ProviderAPIs

_build_resource_relation = json_home_relations.os_oauth2_resource_rel_func


class AccessTokenResource(ks_flask.ResourceBase):
    def _method_not_allowed(self):
        """Raise a method not allowed error."""
        raise exception.OAuth2OtherError(
            int(http.client.METHOD_NOT_ALLOWED),
            http.client.responses[http.client.METHOD_NOT_ALLOWED],
            _('The method is not allowed for the requested URL.'),
        )

    @ks_flask.unenforced_api
    def get(self):
        """The method is not allowed."""
        self._method_not_allowed()

    @ks_flask.unenforced_api
    def head(self):
        """The method is not allowed."""
        self._method_not_allowed()

    @ks_flask.unenforced_api
    def put(self):
        """The method is not allowed."""
        self._method_not_allowed()

    @ks_flask.unenforced_api
    def patch(self):
        """The method is not allowed."""
        self._method_not_allowed()

    @ks_flask.unenforced_api
    def delete(self):
        """The method is not allowed."""
        self._method_not_allowed()

    @ks_flask.unenforced_api
    def post(self):
        """Get an OAuth2.0 Access Token.

        POST /v3/OS-OAUTH2/token
        """
        grant_type = flask.request.form.get('grant_type')
        if grant_type is None:
            error = exception.OAuth2InvalidRequest(
                int(http.client.BAD_REQUEST),
                http.client.responses[http.client.BAD_REQUEST],
                _('The parameter grant_type is required.'),
            )
            LOG.info(
                'Get OAuth2.0 Access Token API: ' f'{error.message_format}'
            )
            raise error
        if grant_type != 'client_credentials':
            error = exception.OAuth2UnsupportedGrantType(
                int(http.client.BAD_REQUEST),
                http.client.responses[http.client.BAD_REQUEST],
                _('The parameter grant_type %s is not supported.')
                % grant_type,
            )
            LOG.info(
                'Get OAuth2.0 Access Token API: ' f'{error.message_format}'
            )
            raise error

        auth_method = ''
        client_id = flask.request.form.get('client_id')
        client_secret = flask.request.form.get('client_secret')
        client_cert = flask.request.environ.get("SSL_CLIENT_CERT")
        client_auth = flask.request.authorization
        if not client_cert and client_auth and client_auth.type == 'basic':
            client_id = client_auth.username
            client_secret = client_auth.password

        if not client_id:
            error = exception.OAuth2InvalidClient(
                int(http.client.UNAUTHORIZED),
                http.client.responses[http.client.UNAUTHORIZED],
                _('Client authentication failed.'),
            )
            LOG.info(
                'Get OAuth2.0 Access Token API: '
                'failed to get a client_id from the request.'
            )
            raise error
        if client_cert:
            auth_method = 'tls_client_auth'
        elif client_secret:
            auth_method = 'client_secret_basic'

        if auth_method in CONF.oauth2.oauth2_authn_methods:
            if auth_method == 'tls_client_auth':
                return self._tls_client_auth(client_id, client_cert)
            if auth_method == 'client_secret_basic':
                return self._client_secret_basic(client_id, client_secret)

        error = exception.OAuth2InvalidClient(
            int(http.client.UNAUTHORIZED),
            http.client.responses[http.client.UNAUTHORIZED],
            _('Client authentication failed.'),
        )
        LOG.info(
            'Get OAuth2.0 Access Token API: '
            'failed to get client credentials from the request.'
        )
        raise error

    def _client_secret_basic(self, client_id, client_secret):
        """Get an OAuth2.0 basic Access Token."""
        auth_data = {
            'identity': {
                'methods': ['application_credential'],
                'application_credential': {
                    'id': client_id,
                    'secret': client_secret,
                },
            }
        }
        try:
            token = authentication.authenticate_for_token(auth_data)
        except exception.Error as error:
            if error.code == 401:
                error = exception.OAuth2InvalidClient(
                    error.code, error.title, str(error)
                )
            elif error.code == 400:
                error = exception.OAuth2InvalidRequest(
                    error.code, error.title, str(error)
                )
            else:
                error = exception.OAuth2OtherError(
                    error.code,
                    error.title,
                    'An unknown error occurred and failed to get an OAuth2.0 '
                    'access token.',
                )
            LOG.exception(error)
            raise error
        except Exception as error:
            error = exception.OAuth2OtherError(
                int(http.client.INTERNAL_SERVER_ERROR),
                http.client.responses[http.client.INTERNAL_SERVER_ERROR],
                str(error),
            )
            LOG.exception(error)
            raise error

        resp = make_response(
            {
                'access_token': token.id,
                'token_type': 'Bearer',
                'expires_in': CONF.token.expiration,
            }
        )
        resp.status = '200 OK'
        return resp

    def _check_mapped_properties(self, cert_dn, user, user_domain):
        mapping_id = CONF.oauth2.get('oauth2_cert_dn_mapping_id')
        try:
            mapping = PROVIDERS.federation_api.get_mapping(mapping_id)
        except exception.MappingNotFound:
            error = exception.OAuth2InvalidClient(
                int(http.client.UNAUTHORIZED),
                http.client.responses[http.client.UNAUTHORIZED],
                _('Client authentication failed.'),
            )
            LOG.info(
                'Get OAuth2.0 Access Token API: '
                'mapping id %s is not found. ',
                mapping_id,
            )
            raise error

        rule_processor = federation_utils.RuleProcessor(
            mapping.get('id'), mapping.get('rules')
        )
        try:
            mapped_properties = rule_processor.process(cert_dn)
        except exception.Error as error:
            LOG.exception(error)
            error = exception.OAuth2InvalidClient(
                int(http.client.UNAUTHORIZED),
                http.client.responses[http.client.UNAUTHORIZED],
                _('Client authentication failed.'),
            )
            LOG.info(
                'Get OAuth2.0 Access Token API: '
                'mapping rule process failed. '
                'mapping_id: %s, rules: %s, data: %s.',
                mapping_id,
                mapping.get('rules'),
                jsonutils.dumps(cert_dn),
            )
            raise error
        except Exception as error:
            LOG.exception(error)
            error = exception.OAuth2OtherError(
                int(http.client.INTERNAL_SERVER_ERROR),
                http.client.responses[http.client.INTERNAL_SERVER_ERROR],
                str(error),
            )
            LOG.info(
                'Get OAuth2.0 Access Token API: '
                'mapping rule process failed. '
                'mapping_id: %s, rules: %s, data: %s.',
                mapping_id,
                mapping.get('rules'),
                jsonutils.dumps(cert_dn),
            )
            raise error

        mapping_user = mapped_properties.get('user', {})
        mapping_user_name = mapping_user.get('name')
        mapping_user_id = mapping_user.get('id')
        mapping_user_email = mapping_user.get('email')
        mapping_domain = mapping_user.get('domain', {})
        mapping_user_domain_id = mapping_domain.get('id')
        mapping_user_domain_name = mapping_domain.get('name')
        if mapping_user_name and mapping_user_name != user.get('name'):
            err = exception.OAuth2InvalidClient(
                int(http.client.UNAUTHORIZED),
                http.client.responses[http.client.UNAUTHORIZED],
                _('Client authentication failed.'),
            )
            LOG.info(
                'Get OAuth2.0 Access Token API: %s check failed. '
                'DN value: %s, DB value: %s.',
                'user name',
                mapping_user_name,
                user.get('name'),
            )
            raise err
        if mapping_user_id and mapping_user_id != user.get('id'):
            err = exception.OAuth2InvalidClient(
                int(http.client.UNAUTHORIZED),
                http.client.responses[http.client.UNAUTHORIZED],
                _('Client authentication failed.'),
            )
            LOG.info(
                'Get OAuth2.0 Access Token API: %s check failed. '
                'DN value: %s, DB value: %s.',
                'user id',
                mapping_user_id,
                user.get('id'),
            )
            raise err
        if mapping_user_email and mapping_user_email != user.get('email'):
            err = exception.OAuth2InvalidClient(
                int(http.client.UNAUTHORIZED),
                http.client.responses[http.client.UNAUTHORIZED],
                _('Client authentication failed.'),
            )
            LOG.info(
                'Get OAuth2.0 Access Token API: %s check failed. '
                'DN value: %s, DB value: %s.',
                'user email',
                mapping_user_email,
                user.get('email'),
            )
            raise err
        if (
            mapping_user_domain_id
            and mapping_user_domain_id != user_domain.get('id')
        ):
            err = exception.OAuth2InvalidClient(
                int(http.client.UNAUTHORIZED),
                http.client.responses[http.client.UNAUTHORIZED],
                _('Client authentication failed.'),
            )
            LOG.info(
                'Get OAuth2.0 Access Token API: %s check failed. '
                'DN value: %s, DB value: %s.',
                'user domain id',
                mapping_user_domain_id,
                user_domain.get('id'),
            )
            raise err
        if (
            mapping_user_domain_name
            and mapping_user_domain_name != user_domain.get('name')
        ):
            err = exception.OAuth2InvalidClient(
                int(http.client.UNAUTHORIZED),
                http.client.responses[http.client.UNAUTHORIZED],
                _('Client authentication failed.'),
            )
            LOG.info(
                'Get OAuth2.0 Access Token API: %s check failed. '
                'DN value: %s, DB value: %s.',
                'user domain name',
                mapping_user_domain_name,
                user_domain.get('name'),
            )
            raise err

    def _tls_client_auth(self, client_id, client_cert):
        """Get an OAuth2.0 certificate-bound Access Token."""
        try:
            cert_subject_dn = utils.get_certificate_subject_dn(client_cert)
        except exception.ValidationError:
            error = exception.OAuth2InvalidClient(
                int(http.client.UNAUTHORIZED),
                http.client.responses[http.client.UNAUTHORIZED],
                _('Client authentication failed.'),
            )
            LOG.info(
                'Get OAuth2.0 Access Token API: '
                'failed to get the subject DN from the certificate.'
            )
            raise error
        try:
            cert_issuer_dn = utils.get_certificate_issuer_dn(client_cert)
        except exception.ValidationError:
            error = exception.OAuth2InvalidClient(
                int(http.client.UNAUTHORIZED),
                http.client.responses[http.client.UNAUTHORIZED],
                _('Client authentication failed.'),
            )
            LOG.info(
                'Get OAuth2.0 Access Token API: '
                'failed to get the issuer DN from the certificate.'
            )
            raise error
        client_cert_dn = {}
        for key in cert_subject_dn:
            client_cert_dn[f'SSL_CLIENT_SUBJECT_DN_{key.upper()}'] = (
                cert_subject_dn.get(key)
            )
        for key in cert_issuer_dn:
            client_cert_dn[f'SSL_CLIENT_ISSUER_DN_{key.upper()}'] = (
                cert_issuer_dn.get(key)
            )

        try:
            user = PROVIDERS.identity_api.get_user(client_id)
        except exception.UserNotFound:
            error = exception.OAuth2InvalidClient(
                int(http.client.UNAUTHORIZED),
                http.client.responses[http.client.UNAUTHORIZED],
                _('Client authentication failed.'),
            )
            LOG.info(
                'Get OAuth2.0 Access Token API: '
                'the user does not exist. user id: %s.',
                client_id,
            )
            raise error
        project_id = user.get('default_project_id')
        if not project_id:
            error = exception.OAuth2InvalidClient(
                int(http.client.UNAUTHORIZED),
                http.client.responses[http.client.UNAUTHORIZED],
                _('Client authentication failed.'),
            )
            LOG.info(
                'Get OAuth2.0 Access Token API: '
                'the user does not have default project. user id: %s.',
                client_id,
            )
            raise error

        user_domain = PROVIDERS.resource_api.get_domain(user.get('domain_id'))
        self._check_mapped_properties(client_cert_dn, user, user_domain)
        thumbprint = utils.get_certificate_thumbprint(client_cert)
        LOG.debug(f'The mTLS certificate thumbprint: {thumbprint}')
        try:
            token = PROVIDERS.token_provider_api.issue_token(
                user_id=client_id,
                method_names=['oauth2_credential'],
                project_id=project_id,
                thumbprint=thumbprint,
            )
        except exception.Error as error:
            if error.code == 401:
                error = exception.OAuth2InvalidClient(
                    error.code, error.title, str(error)
                )
            elif error.code == 400:
                error = exception.OAuth2InvalidRequest(
                    error.code, error.title, str(error)
                )
            else:
                error = exception.OAuth2OtherError(
                    error.code,
                    error.title,
                    'An unknown error occurred and failed to get an OAuth2.0 '
                    'access token.',
                )
            LOG.exception(error)
            raise error
        except Exception as error:
            error = exception.OAuth2OtherError(
                int(http.client.INTERNAL_SERVER_ERROR),
                http.client.responses[http.client.INTERNAL_SERVER_ERROR],
                str(error),
            )
            LOG.exception(error)
            raise error

        resp = make_response(
            {
                'access_token': token.id,
                'token_type': 'Bearer',
                'expires_in': CONF.token.expiration,
            }
        )
        resp.status = '200 OK'
        return resp


class OSAuth2API(ks_flask.APIBase):
    _name = 'OS-OAUTH2'
    _import_name = __name__
    _api_url_prefix = '/OS-OAUTH2'

    resource_mapping = [
        ks_flask.construct_resource_map(
            resource=AccessTokenResource,
            url='/token',
            rel='token',
            resource_kwargs={},
            resource_relation_func=_build_resource_relation,
        )
    ]


APIs = (OSAuth2API,)



# Copyright 2018 Catalyst Cloud Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from keystone.conf import utils

key_repository = cfg.StrOpt(
    'key_repository',
    default='/etc/keystone/fernet-keys/',
    help=utils.fmt(
        """
Directory containing Fernet receipt keys. This directory must exist before
using `keystone-manage fernet_setup` for the first time, must be writable by
the user running `keystone-manage fernet_setup` or `keystone-manage
fernet_rotate`, and of course must be readable by keystone's server process.
The repository may contain keys in one of three states: a single staged key
(always index 0) used for receipt validation, a single primary key (always the
highest index) used for receipt creation and validation, and any number of
secondary keys (all other index values) used for receipt validation. With
multiple keystone nodes, each node must share the same key repository contents,
with the exception of the staged key (index 0). It is safe to run
`keystone-manage fernet_rotate` once on any one node to promote a staged key
(index 0) to be the new primary (incremented from the previous highest index),
and produce a new staged key (a new key with index 0); the resulting repository
can then be atomically replicated to other nodes without any risk of race
conditions (for example, it is safe to run `keystone-manage fernet_rotate` on
host A, wait any amount of time, create a tarball of the directory on host A,
unpack it on host B to a temporary location, and atomically move (`mv`) the
directory into place on host B). Running `keystone-manage fernet_rotate`
*twice* on a key repository without syncing other nodes will result in receipts
that can not be validated by all nodes.
"""
    ),
)

max_active_keys = cfg.IntOpt(
    'max_active_keys',
    default=3,
    min=1,
    help=utils.fmt(
        """
This controls how many keys are held in rotation by `keystone-manage
fernet_rotate` before they are discarded. The default value of 3 means that
keystone will maintain one staged key (always index 0), one primary key (the
highest numerical index), and one secondary key (every other index). Increasing
this value means that additional secondary keys will be kept in the rotation.
"""
    ),
)


GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [key_repository, max_active_keys]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from keystone.conf import utils

driver = cfg.StrOpt(
    'driver',
    default='sql',
    help=utils.fmt(
        """
Entry point for the endpoint filter driver in the
`keystone.endpoint_filter` namespace. Only a `sql` option is provided by
keystone, so there is no reason to set this unless you are providing a custom
entry point.
"""
    ),
)

return_all_endpoints_if_no_filter = cfg.BoolOpt(
    'return_all_endpoints_if_no_filter',
    default=True,
    help=utils.fmt(
        """
This controls keystone's behavior if the configured endpoint filters do not
result in any endpoints for a user + project pair (and therefore a potentially
empty service catalog). If set to true, keystone will return the entire service
catalog. If set to false, keystone will return an empty service catalog.
"""
    ),
)


GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [driver, return_all_endpoints_if_no_filter]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives import serialization


def create_jws_keypair(private_key_path, public_key_path):
    """Create an ECDSA key pair using an secp256r1, or NIST P-256, curve.

    :param private_key_path: location to save the private key
    :param public_key_path: location to save the public key

    """
    private_key = ec.generate_private_key(ec.SECP256R1(), default_backend())

    with open(private_key_path, 'wb') as f:
        f.write(
            private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )

    public_key = private_key.public_key()
    with open(public_key_path, 'wb') as f:
        f.write(
            public_key.public_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PublicFormat.SubjectPublicKeyInfo,
            )
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys

from oslo_config import cfg
from oslo_log import versionutils

from keystone.conf import utils

expiration = cfg.IntOpt(
    'expiration',
    default=3600,
    min=0,
    max=sys.maxsize,
    help=utils.fmt(
        """
The amount of time that a token should remain valid (in seconds). Drastically
reducing this value may break "long-running" operations that involve multiple
services to coordinate together, and will force users to authenticate with
keystone more frequently. Drastically increasing this value will increase the
number of tokens that will be simultaneously valid. Keystone tokens are also
bearer tokens, so a shorter duration will also reduce the potential security
impact of a compromised token.
"""
    ),
)

provider = cfg.StrOpt(
    'provider',
    default='fernet',
    help=utils.fmt(
        """
Entry point for the token provider in the `keystone.token.provider` namespace.
The token provider controls the token construction, validation, and revocation
operations. Supported upstream providers are `fernet` and `jws`. Neither
`fernet` or `jws` tokens require persistence and both require additional setup.
If using `fernet`, you're required to run `keystone-manage fernet_setup`, which
creates symmetric keys used to encrypt tokens. If using `jws`, you're required
to generate an ECDSA keypair using a SHA-256 hash algorithm for signing and
validating token, which can be done with `keystone-manage create_jws_keypair`.
Note that `fernet` tokens are encrypted and `jws` tokens are only signed.
Please be sure to consider this if your deployment has security requirements
regarding payload contents used to generate token IDs.
"""
    ),
)

caching = cfg.BoolOpt(
    'caching',
    default=True,
    help=utils.fmt(
        """
Toggle for caching token creation and validation data. This has no effect
unless global caching is enabled.
"""
    ),
)

cache_time = cfg.IntOpt(
    'cache_time',
    min=0,
    max=sys.maxsize,
    help=utils.fmt(
        """
The number of seconds to cache token creation and validation data. This has no
effect unless both global and `[token] caching` are enabled.
"""
    ),
)

revoke_by_id = cfg.BoolOpt(
    'revoke_by_id',
    default=True,
    help=utils.fmt(
        """
This toggles support for revoking individual tokens by the token identifier and
thus various token enumeration operations (such as listing all tokens issued to
a specific user). These operations are used to determine the list of tokens to
consider revoked. Do not disable this option if you're using the `kvs`
`[revoke] driver`.
"""
    ),
)

allow_rescope_scoped_token = cfg.BoolOpt(
    'allow_rescope_scoped_token',
    default=True,
    help=utils.fmt(
        """
This toggles whether scoped tokens may be re-scoped to a new project or
domain, thereby preventing users from exchanging a scoped token (including
those with a default project scope) for any other token. This forces users to
either authenticate for unscoped tokens (and later exchange that unscoped token
for tokens with a more specific scope) or to provide their credentials in every
request for a scoped token to avoid re-scoping altogether.
"""
    ),
)

cache_on_issue = cfg.BoolOpt(
    'cache_on_issue',
    default=True,
    deprecated_since=versionutils.deprecated.STEIN,
    deprecated_reason=utils.fmt(
        """
Keystone already exposes a configuration option for caching tokens. Having a
separate configuration option to cache tokens when they are issued is
redundant, unnecessarily complicated, and is misleading if token caching is
disabled because tokens will still be pre-cached by default when they are
issued. The ability to pre-cache tokens when they are issued is going to rely
exclusively on the ``keystone.conf [token] caching`` option in the future.
"""
    ),
    deprecated_for_removal=True,
    help=utils.fmt(
        """
Enable storing issued token data to token validation cache so that first token
validation doesn't actually cause full validation cycle. This option has no
effect unless global caching is enabled and will still cache tokens even if
`[token] caching = False`.
"""
    ),
)

allow_expired_window = cfg.IntOpt(
    'allow_expired_window',
    default=48 * 60 * 60,
    help=utils.fmt(
        """
This controls the number of seconds that a token can be retrieved for beyond
the built-in expiry time. This allows long running operations to succeed.
Defaults to two days.
"""
    ),
)


GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [
    expiration,
    provider,
    caching,
    cache_time,
    revoke_by_id,
    allow_rescope_scoped_token,
    cache_on_issue,
    allow_expired_window,
]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
from unittest import mock
import uuid

from testtools import matchers

from keystone.common import driver_hints
from keystone.common import provider_api
from keystone.common.resource_options import options as ro_opt
import keystone.conf
from keystone import exception
from keystone.resource.backends import sql as resource_sql
from keystone.tests import unit
from keystone.tests.unit import default_fixtures
from keystone.tests.unit import utils as test_utils

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class ResourceTests:
    domain_count = len(default_fixtures.DOMAINS)

    def test_get_project(self):
        project_ref = PROVIDERS.resource_api.get_project(
            self.project_bar['id']
        )
        self.assertDictEqual(self.project_bar, project_ref)

    def test_get_project_returns_not_found(self):
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.get_project,
            uuid.uuid4().hex,
        )

    def test_get_project_by_name(self):
        project_ref = PROVIDERS.resource_api.get_project_by_name(
            self.project_bar['name'], CONF.identity.default_domain_id
        )
        self.assertDictEqual(self.project_bar, project_ref)

    @unit.skip_if_no_multiple_domains_support
    def test_get_project_by_name_for_project_acting_as_a_domain(self):
        """Test get_project_by_name works when the domain_id is None."""
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id, is_domain=False
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)

        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.get_project_by_name,
            project['name'],
            None,
        )

        # Test that querying with domain_id as None will find the project
        # acting as a domain, even if it's name is the same as the regular
        # project above.
        project2 = unit.new_project_ref(is_domain=True, name=project['name'])
        project2 = PROVIDERS.resource_api.create_project(
            project2['id'], project2
        )

        project_ref = PROVIDERS.resource_api.get_project_by_name(
            project2['name'], None
        )

        self.assertEqual(project2, project_ref)

    def test_get_project_by_name_returns_not_found(self):
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.get_project_by_name,
            uuid.uuid4().hex,
            CONF.identity.default_domain_id,
        )

    def test_create_duplicate_project_id_fails(self):
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project_id = project['id']
        PROVIDERS.resource_api.create_project(project_id, project)
        project['name'] = 'fake2'
        self.assertRaises(
            exception.Conflict,
            PROVIDERS.resource_api.create_project,
            project_id,
            project,
        )

    def test_create_duplicate_project_name_fails(self):
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project_id = project['id']
        PROVIDERS.resource_api.create_project(project_id, project)
        project['id'] = 'fake2'
        self.assertRaises(
            exception.Conflict,
            PROVIDERS.resource_api.create_project,
            project['id'],
            project,
        )

    def test_create_project_name_with_trailing_whitespace(self):
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project_id = project['id']
        project_name = project['name'] = project['name'] + '    '
        project_returned = PROVIDERS.resource_api.create_project(
            project_id, project
        )
        self.assertEqual(project_id, project_returned['id'])
        self.assertEqual(project_name.strip(), project_returned['name'])

    def test_create_duplicate_project_name_in_different_domains(self):
        new_domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(new_domain['id'], new_domain)
        project1 = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project2 = unit.new_project_ref(
            name=project1['name'], domain_id=new_domain['id']
        )
        PROVIDERS.resource_api.create_project(project1['id'], project1)
        PROVIDERS.resource_api.create_project(project2['id'], project2)

    def test_rename_duplicate_project_name_fails(self):
        project1 = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project2 = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        PROVIDERS.resource_api.create_project(project1['id'], project1)
        PROVIDERS.resource_api.create_project(project2['id'], project2)
        project2['name'] = project1['name']
        self.assertRaises(
            exception.Error,
            PROVIDERS.resource_api.update_project,
            project2['id'],
            project2,
        )

    def test_update_project_id_does_nothing(self):
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project_id = project['id']
        PROVIDERS.resource_api.create_project(project['id'], project)
        project['id'] = 'fake2'
        PROVIDERS.resource_api.update_project(project_id, project)
        project_ref = PROVIDERS.resource_api.get_project(project_id)
        self.assertEqual(project_id, project_ref['id'])
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.get_project,
            'fake2',
        )

    def test_update_project_name_with_trailing_whitespace(self):
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project_id = project['id']
        project_create = PROVIDERS.resource_api.create_project(
            project_id, project
        )
        self.assertEqual(project_id, project_create['id'])
        project_name = project['name'] = project['name'] + '    '
        project_update = PROVIDERS.resource_api.update_project(
            project_id, project
        )
        self.assertEqual(project_id, project_update['id'])
        self.assertEqual(project_name.strip(), project_update['name'])

    def test_delete_domain_with_user_group_project_links(self):
        # TODO(chungg):add test case once expected behaviour defined
        pass

    def test_update_project_returns_not_found(self):
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.update_project,
            uuid.uuid4().hex,
            {},
        )

    def test_delete_project_returns_not_found(self):
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.delete_project,
            uuid.uuid4().hex,
        )

    def test_create_update_delete_unicode_project(self):
        unicode_project_name = 'name \u540d\u5b57'
        project = unit.new_project_ref(
            name=unicode_project_name,
            domain_id=CONF.identity.default_domain_id,
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.resource_api.update_project(project['id'], project)
        PROVIDERS.resource_api.delete_project(project['id'])

    def test_create_project_with_no_enabled_field(self):
        ref = unit.new_project_ref(domain_id=CONF.identity.default_domain_id)
        del ref['enabled']
        PROVIDERS.resource_api.create_project(ref['id'], ref)

        project = PROVIDERS.resource_api.get_project(ref['id'])
        self.assertIs(project['enabled'], True)

    def test_create_project_long_name_fails(self):
        project = unit.new_project_ref(
            name='a' * 65, domain_id=CONF.identity.default_domain_id
        )
        self.assertRaises(
            exception.ValidationError,
            PROVIDERS.resource_api.create_project,
            project['id'],
            project,
        )

    def test_create_project_invalid_domain_id(self):
        project = unit.new_project_ref(domain_id=uuid.uuid4().hex)
        self.assertRaises(
            exception.DomainNotFound,
            PROVIDERS.resource_api.create_project,
            project['id'],
            project,
        )

    def test_list_domains(self):
        domain1 = unit.new_domain_ref()
        domain2 = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain1['id'], domain1)
        PROVIDERS.resource_api.create_domain(domain2['id'], domain2)
        domains = PROVIDERS.resource_api.list_domains()
        self.assertEqual(3, len(domains))
        domain_ids = []
        for domain in domains:
            domain_ids.append(domain.get('id'))
        self.assertIn(CONF.identity.default_domain_id, domain_ids)
        self.assertIn(domain1['id'], domain_ids)
        self.assertIn(domain2['id'], domain_ids)

    def test_list_projects(self):
        project_refs = PROVIDERS.resource_api.list_projects()
        project_count = len(default_fixtures.PROJECTS) + self.domain_count
        self.assertEqual(project_count, len(project_refs))
        for project in default_fixtures.PROJECTS:
            self.assertIn(project, project_refs)

    def test_list_projects_with_multiple_filters(self):
        # Create a project
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)

        # Build driver hints with the project's name and inexistent description
        hints = driver_hints.Hints()
        hints.add_filter('name', project['name'])
        hints.add_filter('description', uuid.uuid4().hex)

        # Retrieve projects based on hints and check an empty list is returned
        projects = PROVIDERS.resource_api.list_projects(hints)
        self.assertEqual([], projects)

        # Build correct driver hints
        hints = driver_hints.Hints()
        hints.add_filter('name', project['name'])
        hints.add_filter('description', project['description'])

        # Retrieve projects based on hints
        projects = PROVIDERS.resource_api.list_projects(hints)

        # Check that the returned list contains only the first project
        self.assertEqual(1, len(projects))
        self.assertEqual(project, projects[0])

    def test_list_projects_for_domain(self):
        project_ids = [
            x['id']
            for x in PROVIDERS.resource_api.list_projects_in_domain(
                CONF.identity.default_domain_id
            )
        ]
        # Only the projects from the default fixtures are expected, since
        # filtering by domain does not include any project that acts as a
        # domain.
        self.assertThat(
            project_ids, matchers.HasLength(len(default_fixtures.PROJECTS))
        )
        self.assertIn(self.project_bar['id'], project_ids)
        self.assertIn(self.project_baz['id'], project_ids)
        self.assertIn(self.project_mtu['id'], project_ids)
        self.assertIn(self.project_service['id'], project_ids)

    @unit.skip_if_no_multiple_domains_support
    def test_list_projects_acting_as_domain(self):
        initial_domains = PROVIDERS.resource_api.list_domains()

        # Creating 5 projects that act as domains
        new_projects_acting_as_domains = []
        for i in range(5):
            project = unit.new_project_ref(is_domain=True)
            project = PROVIDERS.resource_api.create_project(
                project['id'], project
            )
            new_projects_acting_as_domains.append(project)

        # Creating a few regular project to ensure it doesn't mess with the
        # ones that act as domains
        self._create_projects_hierarchy(hierarchy_size=2)

        projects = PROVIDERS.resource_api.list_projects_acting_as_domain()
        expected_number_projects = len(initial_domains) + len(
            new_projects_acting_as_domains
        )
        self.assertEqual(expected_number_projects, len(projects))
        for project in new_projects_acting_as_domains:
            self.assertIn(project, projects)
        for domain in initial_domains:
            self.assertIn(domain['id'], [p['id'] for p in projects])

    @unit.skip_if_no_multiple_domains_support
    def test_list_projects_for_alternate_domain(self):
        domain1 = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain1['id'], domain1)
        project1 = unit.new_project_ref(domain_id=domain1['id'])
        PROVIDERS.resource_api.create_project(project1['id'], project1)
        project2 = unit.new_project_ref(domain_id=domain1['id'])
        PROVIDERS.resource_api.create_project(project2['id'], project2)
        project_ids = [
            x['id']
            for x in PROVIDERS.resource_api.list_projects_in_domain(
                domain1['id']
            )
        ]
        self.assertEqual(2, len(project_ids))
        self.assertIn(project1['id'], project_ids)
        self.assertIn(project2['id'], project_ids)

    def _create_projects_hierarchy(
        self,
        hierarchy_size=2,
        domain_id=None,
        is_domain=False,
        parent_project_id=None,
    ):
        """Create a project hierarchy with specified size.

        :param hierarchy_size: the desired hierarchy size, default is 2 -
                               a project with one child.
        :param domain_id: domain where the projects hierarchy will be created.
        :param is_domain: if the hierarchy will have the is_domain flag active
                          or not.
        :param parent_project_id: if the intention is to create a
            sub-hierarchy, sets the sub-hierarchy root. Defaults to creating
            a new hierarchy, i.e. a new root project.

        :returns projects: a list of the projects in the created hierarchy.

        """
        if domain_id is None:
            domain_id = CONF.identity.default_domain_id
        if parent_project_id:
            project = unit.new_project_ref(
                parent_id=parent_project_id,
                domain_id=domain_id,
                is_domain=is_domain,
            )
        else:
            project = unit.new_project_ref(
                domain_id=domain_id, is_domain=is_domain
            )
        project_id = project['id']
        project = PROVIDERS.resource_api.create_project(project_id, project)

        projects = [project]
        for i in range(1, hierarchy_size):
            new_project = unit.new_project_ref(
                parent_id=project_id, domain_id=domain_id
            )

            PROVIDERS.resource_api.create_project(
                new_project['id'], new_project
            )
            projects.append(new_project)
            project_id = new_project['id']

        return projects

    @unit.skip_if_no_multiple_domains_support
    def test_create_domain_with_project_api(self):
        project = unit.new_project_ref(is_domain=True)
        ref = PROVIDERS.resource_api.create_project(project['id'], project)
        self.assertTrue(ref['is_domain'])
        PROVIDERS.resource_api.get_domain(ref['id'])

    @unit.skip_if_no_multiple_domains_support
    def test_project_as_a_domain_uniqueness_constraints(self):
        """Test project uniqueness for those acting as domains.

        If it is a project acting as a domain, we can't have two or more with
        the same name.

        """
        # Create two projects acting as a domain
        project = unit.new_project_ref(is_domain=True)
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        project2 = unit.new_project_ref(is_domain=True)
        project2 = PROVIDERS.resource_api.create_project(
            project2['id'], project2
        )

        # All projects acting as domains have a null domain_id, so should not
        # be able to create another with the same name but a different
        # project ID.
        new_project = project.copy()
        new_project['id'] = uuid.uuid4().hex

        self.assertRaises(
            exception.Conflict,
            PROVIDERS.resource_api.create_project,
            new_project['id'],
            new_project,
        )

        # We also should not be able to update one to have a name clash
        project2['name'] = project['name']
        self.assertRaises(
            exception.Conflict,
            PROVIDERS.resource_api.update_project,
            project2['id'],
            project2,
        )

        # But updating it to a unique name is OK
        project2['name'] = uuid.uuid4().hex
        PROVIDERS.resource_api.update_project(project2['id'], project2)

        # Finally, it should be OK to create a project with same name as one of
        # these acting as a domain, as long as it is a regular project
        project3 = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id, name=project2['name']
        )
        PROVIDERS.resource_api.create_project(project3['id'], project3)
        # In fact, it should be OK to create such a project in the domain which
        # has the matching name.
        # TODO(henry-nash): Once we fully support projects acting as a domain,
        # add a test here to create a sub-project with a name that matches its
        # project acting as a domain

    @unit.skip_if_no_multiple_domains_support
    @test_utils.wip('waiting for sub projects acting as domains support')
    def test_is_domain_sub_project_has_parent_domain_id(self):
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id, is_domain=True
        )
        PROVIDERS.resource_api.create_project(project['id'], project)

        sub_project = unit.new_project_ref(
            domain_id=project['id'], parent_id=project['id'], is_domain=True
        )

        ref = PROVIDERS.resource_api.create_project(
            sub_project['id'], sub_project
        )
        self.assertTrue(ref['is_domain'])
        self.assertEqual(project['id'], ref['parent_id'])
        self.assertEqual(project['id'], ref['domain_id'])

    @unit.skip_if_no_multiple_domains_support
    def test_delete_domain_with_project_api(self):
        project = unit.new_project_ref(domain_id=None, is_domain=True)
        PROVIDERS.resource_api.create_project(project['id'], project)

        # Check that a corresponding domain was created
        PROVIDERS.resource_api.get_domain(project['id'])

        # Try to delete the enabled project that acts as a domain
        self.assertRaises(
            exception.ForbiddenNotSecurity,
            PROVIDERS.resource_api.delete_project,
            project['id'],
        )

        # Disable the project
        project['enabled'] = False
        PROVIDERS.resource_api.update_project(project['id'], project)

        # Successfully delete the project
        PROVIDERS.resource_api.delete_project(project['id'])

        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.get_project,
            project['id'],
        )

        self.assertRaises(
            exception.DomainNotFound,
            PROVIDERS.resource_api.get_domain,
            project['id'],
        )

    @unit.skip_if_no_multiple_domains_support
    def test_create_subproject_acting_as_domain_fails(self):
        root_project = unit.new_project_ref(is_domain=True)
        PROVIDERS.resource_api.create_project(root_project['id'], root_project)

        sub_project = unit.new_project_ref(
            is_domain=True, parent_id=root_project['id']
        )

        # Creation of sub projects acting as domains is not allowed yet
        self.assertRaises(
            exception.ValidationError,
            PROVIDERS.resource_api.create_project,
            sub_project['id'],
            sub_project,
        )

    @unit.skip_if_no_multiple_domains_support
    def test_create_domain_under_regular_project_hierarchy_fails(self):
        # Projects acting as domains can't have a regular project as parent
        projects_hierarchy = self._create_projects_hierarchy()
        parent = projects_hierarchy[1]
        project = unit.new_project_ref(
            domain_id=parent['id'], parent_id=parent['id'], is_domain=True
        )

        self.assertRaises(
            exception.ValidationError,
            PROVIDERS.resource_api.create_project,
            project['id'],
            project,
        )

    @unit.skip_if_no_multiple_domains_support
    @test_utils.wip('waiting for sub projects acting as domains support')
    def test_create_project_under_domain_hierarchy(self):
        projects_hierarchy = self._create_projects_hierarchy(is_domain=True)
        parent = projects_hierarchy[1]
        project = unit.new_project_ref(
            domain_id=parent['id'], parent_id=parent['id'], is_domain=False
        )

        ref = PROVIDERS.resource_api.create_project(project['id'], project)
        self.assertFalse(ref['is_domain'])
        self.assertEqual(parent['id'], ref['parent_id'])
        self.assertEqual(parent['id'], ref['domain_id'])

    def test_create_project_without_is_domain_flag(self):
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        del project['is_domain']
        ref = PROVIDERS.resource_api.create_project(project['id'], project)
        # The is_domain flag should be False by default
        self.assertFalse(ref['is_domain'])

    @unit.skip_if_no_multiple_domains_support
    def test_create_project_passing_is_domain_flag_true(self):
        project = unit.new_project_ref(is_domain=True)

        ref = PROVIDERS.resource_api.create_project(project['id'], project)
        self.assertTrue(ref['is_domain'])

    def test_create_project_passing_is_domain_flag_false(self):
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id, is_domain=False
        )

        ref = PROVIDERS.resource_api.create_project(project['id'], project)
        self.assertIs(False, ref['is_domain'])

    @test_utils.wip('waiting for support for parent_id to imply domain_id')
    def test_create_project_with_parent_id_and_without_domain_id(self):
        # First create a domain
        project = unit.new_project_ref(is_domain=True)
        PROVIDERS.resource_api.create_project(project['id'], project)
        # Now create a child by just naming the parent_id
        sub_project = unit.new_project_ref(parent_id=project['id'])
        ref = PROVIDERS.resource_api.create_project(
            sub_project['id'], sub_project
        )

        # The domain_id should be set to the parent domain_id
        self.assertEqual(project['domain_id'], ref['domain_id'])

    def test_create_project_with_domain_id_and_without_parent_id(self):
        # First create a domain
        project = unit.new_project_ref(is_domain=True)
        PROVIDERS.resource_api.create_project(project['id'], project)
        # Now create a child by just naming the domain_id
        sub_project = unit.new_project_ref(domain_id=project['id'])
        ref = PROVIDERS.resource_api.create_project(
            sub_project['id'], sub_project
        )

        # The parent_id and domain_id should be set to the id of the project
        # acting as a domain
        self.assertEqual(project['id'], ref['parent_id'])
        self.assertEqual(project['id'], ref['domain_id'])

    def test_create_project_with_domain_id_mismatch_to_parent_domain(self):
        # First create a domain
        project = unit.new_project_ref(is_domain=True)
        PROVIDERS.resource_api.create_project(project['id'], project)
        # Now try to create a child with the above as its parent, but
        # specifying a different domain.
        sub_project = unit.new_project_ref(
            parent_id=project['id'], domain_id=CONF.identity.default_domain_id
        )
        self.assertRaises(
            exception.ValidationError,
            PROVIDERS.resource_api.create_project,
            sub_project['id'],
            sub_project,
        )

    def test_check_leaf_projects(self):
        projects_hierarchy = self._create_projects_hierarchy()
        root_project = projects_hierarchy[0]
        leaf_project = projects_hierarchy[1]

        self.assertFalse(
            PROVIDERS.resource_api.is_leaf_project(root_project['id'])
        )
        self.assertTrue(
            PROVIDERS.resource_api.is_leaf_project(leaf_project['id'])
        )

        # Delete leaf_project
        PROVIDERS.resource_api.delete_project(leaf_project['id'])

        # Now, root_project should be leaf
        self.assertTrue(
            PROVIDERS.resource_api.is_leaf_project(root_project['id'])
        )

    def test_list_projects_in_subtree(self):
        projects_hierarchy = self._create_projects_hierarchy(hierarchy_size=3)
        project1 = projects_hierarchy[0]
        project2 = projects_hierarchy[1]
        project3 = projects_hierarchy[2]
        project4 = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id, parent_id=project2['id']
        )
        PROVIDERS.resource_api.create_project(project4['id'], project4)

        subtree = PROVIDERS.resource_api.list_projects_in_subtree(
            project1['id']
        )
        self.assertEqual(3, len(subtree))
        self.assertIn(project2, subtree)
        self.assertIn(project3, subtree)
        self.assertIn(project4, subtree)

        subtree = PROVIDERS.resource_api.list_projects_in_subtree(
            project2['id']
        )
        self.assertEqual(2, len(subtree))
        self.assertIn(project3, subtree)
        self.assertIn(project4, subtree)

        subtree = PROVIDERS.resource_api.list_projects_in_subtree(
            project3['id']
        )
        self.assertEqual(0, len(subtree))

    def test_get_projects_in_subtree_as_ids_with_large_tree(self):
        """Check project hierarchy is returned correctly in large tree.

        With a large hierarchy we need to enforce the projects
        are returned in the correct order (illustrated below).

        Tree we will create::

               +------p1------+
               |              |
            +---p3---+      +-p2-+
            |        |      |    |
            p7    +-p6-+   p5    p4
            |     |    |
            p10   p9   p8
                  |
                 p11
        """
        # Create large project hierarchy, of above depiction
        p1, p2, p4 = self._create_projects_hierarchy(hierarchy_size=3)
        p5 = self._create_projects_hierarchy(
            hierarchy_size=1, parent_project_id=p2['id']
        )[0]
        p3, p6, p8 = self._create_projects_hierarchy(
            hierarchy_size=3, parent_project_id=p1['id']
        )
        p9, p11 = self._create_projects_hierarchy(
            hierarchy_size=2, parent_project_id=p6['id']
        )
        p7, p10 = self._create_projects_hierarchy(
            hierarchy_size=2, parent_project_id=p3['id']
        )

        expected_projects = {
            p2['id']: {p5['id']: None, p4['id']: None},
            p3['id']: {
                p7['id']: {p10['id']: None},
                p6['id']: {p9['id']: {p11['id']: None}, p8['id']: None},
            },
        }

        prjs_hierarchy = PROVIDERS.resource_api.get_projects_in_subtree_as_ids(
            p1['id']
        )

        self.assertDictEqual(expected_projects, prjs_hierarchy)

    def test_list_projects_in_subtree_with_circular_reference(self):
        project1 = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project1 = PROVIDERS.resource_api.create_project(
            project1['id'], project1
        )

        project2 = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id, parent_id=project1['id']
        )
        PROVIDERS.resource_api.create_project(project2['id'], project2)

        project1['parent_id'] = project2['id']  # Adds cyclic reference

        # NOTE(dstanek): The manager does not allow parent_id to be updated.
        # Instead will directly use the driver to create the cyclic
        # reference.
        PROVIDERS.resource_api.driver.update_project(project1['id'], project1)

        subtree = PROVIDERS.resource_api.list_projects_in_subtree(
            project1['id']
        )

        # NOTE(dstanek): If a cyclic reference is detected the code bails
        # and returns None instead of falling into the infinite
        # recursion trap.
        self.assertIsNone(subtree)

    def test_list_projects_in_subtree_invalid_project_id(self):
        self.assertRaises(
            exception.ValidationError,
            PROVIDERS.resource_api.list_projects_in_subtree,
            None,
        )

        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.list_projects_in_subtree,
            uuid.uuid4().hex,
        )

    def test_list_project_parents(self):
        projects_hierarchy = self._create_projects_hierarchy(hierarchy_size=3)
        project1 = projects_hierarchy[0]
        project2 = projects_hierarchy[1]
        project3 = projects_hierarchy[2]
        project4 = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id, parent_id=project2['id']
        )
        PROVIDERS.resource_api.create_project(project4['id'], project4)

        parents1 = PROVIDERS.resource_api.list_project_parents(project3['id'])
        self.assertEqual(3, len(parents1))
        self.assertIn(project1, parents1)
        self.assertIn(project2, parents1)

        parents2 = PROVIDERS.resource_api.list_project_parents(project4['id'])
        self.assertEqual(parents1, parents2)

        parents = PROVIDERS.resource_api.list_project_parents(project1['id'])
        # It has the default domain as parent
        self.assertEqual(1, len(parents))

    def test_update_project_enabled_cascade(self):
        """Test update_project_cascade.

        Ensures the enabled attribute is correctly updated across
        a simple 3-level projects hierarchy.
        """
        projects_hierarchy = self._create_projects_hierarchy(hierarchy_size=3)
        parent = projects_hierarchy[0]

        # Disable in parent project disables the whole subtree
        parent['enabled'] = False
        # Store the ref from backend in another variable so we don't bother
        # to remove other attributes that were not originally provided and
        # were set in the manager, like parent_id and domain_id.
        parent_ref = PROVIDERS.resource_api.update_project(
            parent['id'], parent, cascade=True
        )

        subtree = PROVIDERS.resource_api.list_projects_in_subtree(parent['id'])
        self.assertEqual(2, len(subtree))
        self.assertFalse(parent_ref['enabled'])
        self.assertFalse(subtree[0]['enabled'])
        self.assertFalse(subtree[1]['enabled'])

        # Enable parent project enables the whole subtree
        parent['enabled'] = True
        parent_ref = PROVIDERS.resource_api.update_project(
            parent['id'], parent, cascade=True
        )

        subtree = PROVIDERS.resource_api.list_projects_in_subtree(parent['id'])
        self.assertEqual(2, len(subtree))
        self.assertTrue(parent_ref['enabled'])
        self.assertTrue(subtree[0]['enabled'])
        self.assertTrue(subtree[1]['enabled'])

    def test_cannot_enable_cascade_with_parent_disabled(self):
        projects_hierarchy = self._create_projects_hierarchy(hierarchy_size=3)
        grandparent = projects_hierarchy[0]
        parent = projects_hierarchy[1]

        grandparent['enabled'] = False
        PROVIDERS.resource_api.update_project(
            grandparent['id'], grandparent, cascade=True
        )
        subtree = PROVIDERS.resource_api.list_projects_in_subtree(parent['id'])
        self.assertFalse(subtree[0]['enabled'])

        parent['enabled'] = True
        self.assertRaises(
            exception.ForbiddenNotSecurity,
            PROVIDERS.resource_api.update_project,
            parent['id'],
            parent,
            cascade=True,
        )

    def test_update_cascade_only_accepts_enabled(self):
        # Update cascade does not accept any other attribute but 'enabled'
        new_project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        PROVIDERS.resource_api.create_project(new_project['id'], new_project)

        new_project['name'] = 'project1'
        self.assertRaises(
            exception.ValidationError,
            PROVIDERS.resource_api.update_project,
            new_project['id'],
            new_project,
            cascade=True,
        )

    def test_list_project_parents_invalid_project_id(self):
        self.assertRaises(
            exception.ValidationError,
            PROVIDERS.resource_api.list_project_parents,
            None,
        )

        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.list_project_parents,
            uuid.uuid4().hex,
        )

    def test_create_project_doesnt_modify_passed_in_dict(self):
        new_project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        original_project = new_project.copy()
        PROVIDERS.resource_api.create_project(new_project['id'], new_project)
        self.assertDictEqual(original_project, new_project)

    def test_update_project_enable(self):
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        PROVIDERS.resource_api.create_project(project['id'], project)
        project_ref = PROVIDERS.resource_api.get_project(project['id'])
        self.assertTrue(project_ref['enabled'])

        project['enabled'] = False
        PROVIDERS.resource_api.update_project(project['id'], project)
        project_ref = PROVIDERS.resource_api.get_project(project['id'])
        self.assertEqual(project['enabled'], project_ref['enabled'])

        # If not present, enabled field should not be updated
        del project['enabled']
        PROVIDERS.resource_api.update_project(project['id'], project)
        project_ref = PROVIDERS.resource_api.get_project(project['id'])
        self.assertFalse(project_ref['enabled'])

        project['enabled'] = True
        PROVIDERS.resource_api.update_project(project['id'], project)
        project_ref = PROVIDERS.resource_api.get_project(project['id'])
        self.assertEqual(project['enabled'], project_ref['enabled'])

        del project['enabled']
        PROVIDERS.resource_api.update_project(project['id'], project)
        project_ref = PROVIDERS.resource_api.get_project(project['id'])
        self.assertTrue(project_ref['enabled'])

    def test_create_invalid_domain_fails(self):
        new_group = unit.new_group_ref(domain_id="doesnotexist")
        self.assertRaises(
            exception.DomainNotFound,
            PROVIDERS.identity_api.create_group,
            new_group,
        )
        new_user = unit.new_user_ref(domain_id="doesnotexist")
        self.assertRaises(
            exception.DomainNotFound,
            PROVIDERS.identity_api.create_user,
            new_user,
        )

    @unit.skip_if_no_multiple_domains_support
    def test_project_crud(self):
        domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain['id'], domain)
        project = unit.new_project_ref(domain_id=domain['id'])
        PROVIDERS.resource_api.create_project(project['id'], project)
        project_ref = PROVIDERS.resource_api.get_project(project['id'])
        self.assertLessEqual(project.items(), project_ref.items())

        project['name'] = uuid.uuid4().hex
        PROVIDERS.resource_api.update_project(project['id'], project)
        project_ref = PROVIDERS.resource_api.get_project(project['id'])
        self.assertLessEqual(project.items(), project_ref.items())

        PROVIDERS.resource_api.delete_project(project['id'])
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.get_project,
            project['id'],
        )

    def test_domain_delete_hierarchy(self):
        domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain['id'], domain)

        # Creating a root and a leaf project inside the domain
        projects_hierarchy = self._create_projects_hierarchy(
            domain_id=domain['id']
        )
        root_project = projects_hierarchy[0]
        leaf_project = projects_hierarchy[0]

        # Disable the domain
        domain['enabled'] = False
        PROVIDERS.resource_api.update_domain(domain['id'], domain)

        # Delete the domain
        PROVIDERS.resource_api.delete_domain(domain['id'])

        # Make sure the domain no longer exists
        self.assertRaises(
            exception.DomainNotFound,
            PROVIDERS.resource_api.get_domain,
            domain['id'],
        )

        # Make sure the root project no longer exists
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.get_project,
            root_project['id'],
        )

        # Make sure the leaf project no longer exists
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.get_project,
            leaf_project['id'],
        )

    def test_delete_projects_from_ids(self):
        """Test the resource backend call delete_projects_from_ids.

        Tests the normal flow of the delete_projects_from_ids backend call,
        that ensures no project on the list exists after it is successfully
        called.
        """
        project1_ref = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project2_ref = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        projects = (project1_ref, project2_ref)
        for project in projects:
            PROVIDERS.resource_api.create_project(project['id'], project)

        # Setting up the ID's list
        projects_ids = [p['id'] for p in projects]
        PROVIDERS.resource_api.driver.delete_projects_from_ids(projects_ids)

        # Ensuring projects no longer exist at backend level
        for project_id in projects_ids:
            self.assertRaises(
                exception.ProjectNotFound,
                PROVIDERS.resource_api.driver.get_project,
                project_id,
            )

        # Passing an empty list is silently ignored
        PROVIDERS.resource_api.driver.delete_projects_from_ids([])

    def test_delete_projects_from_ids_with_no_existing_project_id(self):
        """Test delete_projects_from_ids issues warning if not found.

        Tests the resource backend call delete_projects_from_ids passing a
        non existing ID in project_ids, which is logged and ignored by
        the backend.
        """
        project_ref = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        PROVIDERS.resource_api.create_project(project_ref['id'], project_ref)

        # Setting up the ID's list
        projects_ids = (project_ref['id'], uuid.uuid4().hex)
        with mock.patch('keystone.resource.backends.sql.LOG') as mock_log:
            PROVIDERS.resource_api.delete_projects_from_ids(projects_ids)
            self.assertTrue(mock_log.warning.called)
        # The existing project was deleted.
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.driver.get_project,
            project_ref['id'],
        )

        # Even if we only have one project, and it does not exist, it returns
        # no error.
        PROVIDERS.resource_api.driver.delete_projects_from_ids(
            [uuid.uuid4().hex]
        )

    def test_delete_project_cascade(self):
        # create a hierarchy with 3 levels
        projects_hierarchy = self._create_projects_hierarchy(hierarchy_size=3)
        root_project = projects_hierarchy[0]
        project1 = projects_hierarchy[1]
        project2 = projects_hierarchy[2]

        # Disabling all projects before attempting to delete
        for project in (project2, project1, root_project):
            project['enabled'] = False
            PROVIDERS.resource_api.update_project(project['id'], project)

        PROVIDERS.resource_api.delete_project(root_project['id'], cascade=True)

        for project in projects_hierarchy:
            self.assertRaises(
                exception.ProjectNotFound,
                PROVIDERS.resource_api.get_project,
                project['id'],
            )

    def test_delete_large_project_cascade(self):
        """Try delete a large project with cascade true.

        Tree we will create::

               +-p1-+
               |    |
              p5    p2
               |    |
              p6  +-p3-+
                  |    |
                  p7   p4
        """
        # create a hierarchy with 4 levels
        projects_hierarchy = self._create_projects_hierarchy(hierarchy_size=4)
        p1 = projects_hierarchy[0]
        # Add the left branch to the hierarchy (p5, p6)
        self._create_projects_hierarchy(
            hierarchy_size=2, parent_project_id=p1['id']
        )
        # Add p7 to the hierarchy
        p3_id = projects_hierarchy[2]['id']
        self._create_projects_hierarchy(
            hierarchy_size=1, parent_project_id=p3_id
        )
        # Reverse the hierarchy to disable the leaf first
        prjs_hierarchy = (
            [p1] + PROVIDERS.resource_api.list_projects_in_subtree(p1['id'])
        )[::-1]

        # Disabling all projects before attempting to delete
        for project in prjs_hierarchy:
            project['enabled'] = False
            PROVIDERS.resource_api.update_project(project['id'], project)

        PROVIDERS.resource_api.delete_project(p1['id'], cascade=True)
        for project in prjs_hierarchy:
            self.assertRaises(
                exception.ProjectNotFound,
                PROVIDERS.resource_api.get_project,
                project['id'],
            )

    def test_cannot_delete_project_cascade_with_enabled_child(self):
        # create a hierarchy with 3 levels
        projects_hierarchy = self._create_projects_hierarchy(hierarchy_size=3)
        root_project = projects_hierarchy[0]
        project1 = projects_hierarchy[1]
        project2 = projects_hierarchy[2]

        project2['enabled'] = False
        PROVIDERS.resource_api.update_project(project2['id'], project2)

        # Cannot cascade delete root_project, since project1 is enabled
        self.assertRaises(
            exception.ForbiddenNotSecurity,
            PROVIDERS.resource_api.delete_project,
            root_project['id'],
            cascade=True,
        )

        # Ensuring no project was deleted, not even project2
        PROVIDERS.resource_api.get_project(root_project['id'])
        PROVIDERS.resource_api.get_project(project1['id'])
        PROVIDERS.resource_api.get_project(project2['id'])

    def test_hierarchical_projects_crud(self):
        # create a hierarchy with just a root project (which is a leaf as well)
        projects_hierarchy = self._create_projects_hierarchy(hierarchy_size=1)
        root_project1 = projects_hierarchy[0]

        # create a hierarchy with one root project and one leaf project
        projects_hierarchy = self._create_projects_hierarchy()
        root_project2 = projects_hierarchy[0]
        leaf_project = projects_hierarchy[1]

        # update description from leaf_project
        leaf_project['description'] = 'new description'
        PROVIDERS.resource_api.update_project(leaf_project['id'], leaf_project)
        proj_ref = PROVIDERS.resource_api.get_project(leaf_project['id'])
        self.assertDictEqual(leaf_project, proj_ref)

        # update the parent_id is not allowed
        leaf_project['parent_id'] = root_project1['id']
        self.assertRaises(
            exception.ForbiddenNotSecurity,
            PROVIDERS.resource_api.update_project,
            leaf_project['id'],
            leaf_project,
        )

        # delete root_project1
        PROVIDERS.resource_api.delete_project(root_project1['id'])
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.get_project,
            root_project1['id'],
        )

        # delete root_project2 is not allowed since it is not a leaf project
        self.assertRaises(
            exception.ForbiddenNotSecurity,
            PROVIDERS.resource_api.delete_project,
            root_project2['id'],
        )

    def test_create_project_with_invalid_parent(self):
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id, parent_id='fake'
        )
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.create_project,
            project['id'],
            project,
        )

    @unit.skip_if_no_multiple_domains_support
    def test_create_leaf_project_with_different_domain(self):
        root_project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        PROVIDERS.resource_api.create_project(root_project['id'], root_project)

        domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain['id'], domain)
        leaf_project = unit.new_project_ref(
            domain_id=domain['id'], parent_id=root_project['id']
        )

        self.assertRaises(
            exception.ValidationError,
            PROVIDERS.resource_api.create_project,
            leaf_project['id'],
            leaf_project,
        )

    def test_delete_hierarchical_leaf_project(self):
        projects_hierarchy = self._create_projects_hierarchy()
        root_project = projects_hierarchy[0]
        leaf_project = projects_hierarchy[1]

        PROVIDERS.resource_api.delete_project(leaf_project['id'])
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.get_project,
            leaf_project['id'],
        )

        PROVIDERS.resource_api.delete_project(root_project['id'])
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.get_project,
            root_project['id'],
        )

    def test_delete_hierarchical_not_leaf_project(self):
        projects_hierarchy = self._create_projects_hierarchy()
        root_project = projects_hierarchy[0]

        self.assertRaises(
            exception.ForbiddenNotSecurity,
            PROVIDERS.resource_api.delete_project,
            root_project['id'],
        )

    def test_update_project_parent(self):
        projects_hierarchy = self._create_projects_hierarchy(hierarchy_size=3)
        project1 = projects_hierarchy[0]
        project2 = projects_hierarchy[1]
        project3 = projects_hierarchy[2]

        # project2 is the parent from project3
        self.assertEqual(project3.get('parent_id'), project2['id'])

        # try to update project3 parent to parent1
        project3['parent_id'] = project1['id']
        self.assertRaises(
            exception.ForbiddenNotSecurity,
            PROVIDERS.resource_api.update_project,
            project3['id'],
            project3,
        )

    def test_create_project_under_disabled_one(self):
        project1 = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id, enabled=False
        )
        PROVIDERS.resource_api.create_project(project1['id'], project1)

        project2 = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id, parent_id=project1['id']
        )

        # It's not possible to create a project under a disabled one in the
        # hierarchy
        self.assertRaises(
            exception.ValidationError,
            PROVIDERS.resource_api.create_project,
            project2['id'],
            project2,
        )

    def test_disable_hierarchical_leaf_project(self):
        projects_hierarchy = self._create_projects_hierarchy()
        leaf_project = projects_hierarchy[1]

        leaf_project['enabled'] = False
        PROVIDERS.resource_api.update_project(leaf_project['id'], leaf_project)

        project_ref = PROVIDERS.resource_api.get_project(leaf_project['id'])
        self.assertEqual(leaf_project['enabled'], project_ref['enabled'])

    def test_disable_hierarchical_not_leaf_project(self):
        projects_hierarchy = self._create_projects_hierarchy()
        root_project = projects_hierarchy[0]

        root_project['enabled'] = False
        self.assertRaises(
            exception.ForbiddenNotSecurity,
            PROVIDERS.resource_api.update_project,
            root_project['id'],
            root_project,
        )

    def test_enable_project_with_disabled_parent(self):
        projects_hierarchy = self._create_projects_hierarchy()
        root_project = projects_hierarchy[0]
        leaf_project = projects_hierarchy[1]

        # Disable leaf and root
        leaf_project['enabled'] = False
        PROVIDERS.resource_api.update_project(leaf_project['id'], leaf_project)
        root_project['enabled'] = False
        PROVIDERS.resource_api.update_project(root_project['id'], root_project)

        # Try to enable the leaf project, it's not possible since it has
        # a disabled parent
        leaf_project['enabled'] = True
        self.assertRaises(
            exception.ForbiddenNotSecurity,
            PROVIDERS.resource_api.update_project,
            leaf_project['id'],
            leaf_project,
        )

    def _get_hierarchy_depth(self, project_id):
        return len(PROVIDERS.resource_api.list_project_parents(project_id)) + 1

    def test_check_hierarchy_depth(self):
        # Should be allowed to have a hierarchy of the max depth specified
        # in the config option plus one (to allow for the additional project
        # acting as a domain after an upgrade)
        projects_hierarchy = self._create_projects_hierarchy(
            CONF.max_project_tree_depth
        )
        leaf_project = projects_hierarchy[CONF.max_project_tree_depth - 1]

        depth = self._get_hierarchy_depth(leaf_project['id'])
        self.assertEqual(CONF.max_project_tree_depth + 1, depth)

        # Creating another project in the hierarchy shouldn't be allowed
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id,
            parent_id=leaf_project['id'],
        )
        self.assertRaises(
            exception.ForbiddenNotSecurity,
            PROVIDERS.resource_api.create_project,
            project['id'],
            project,
        )

    def test_project_update_missing_attrs_with_a_value(self):
        # Creating a project with no description attribute.
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        del project['description']
        project = PROVIDERS.resource_api.create_project(project['id'], project)

        # Add a description attribute.
        project['description'] = uuid.uuid4().hex
        PROVIDERS.resource_api.update_project(project['id'], project)

        project_ref = PROVIDERS.resource_api.get_project(project['id'])
        self.assertDictEqual(project, project_ref)

    def test_project_update_missing_attrs_with_a_falsey_value(self):
        # Creating a project with no description attribute.
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        del project['description']
        project = PROVIDERS.resource_api.create_project(project['id'], project)

        # Add a description attribute.
        project['description'] = ''
        PROVIDERS.resource_api.update_project(project['id'], project)

        project_ref = PROVIDERS.resource_api.get_project(project['id'])
        self.assertDictEqual(project, project_ref)

    def test_domain_crud(self):
        domain = unit.new_domain_ref()
        domain_ref = PROVIDERS.resource_api.create_domain(domain['id'], domain)
        self.assertDictEqual(domain, domain_ref)
        domain_ref = PROVIDERS.resource_api.get_domain(domain['id'])
        self.assertDictEqual(domain, domain_ref)

        domain['name'] = uuid.uuid4().hex
        domain_ref = PROVIDERS.resource_api.update_domain(domain['id'], domain)
        self.assertDictEqual(domain, domain_ref)
        domain_ref = PROVIDERS.resource_api.get_domain(domain['id'])
        self.assertDictEqual(domain, domain_ref)

        # Ensure an 'enabled' domain cannot be deleted
        self.assertRaises(
            exception.ForbiddenNotSecurity,
            PROVIDERS.resource_api.delete_domain,
            domain_id=domain['id'],
        )

        # Disable the domain
        domain['enabled'] = False
        PROVIDERS.resource_api.update_domain(domain['id'], domain)

        # Delete the domain
        PROVIDERS.resource_api.delete_domain(domain['id'])

        # Make sure the domain no longer exists
        self.assertRaises(
            exception.DomainNotFound,
            PROVIDERS.resource_api.get_domain,
            domain['id'],
        )

    @unit.skip_if_no_multiple_domains_support
    def test_delete_domain_call_db_time(self):
        domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain['id'], domain)
        # Disable the domain
        domain['enabled'] = False
        PROVIDERS.resource_api.update_domain(domain['id'], domain)

        domain_ref = PROVIDERS.resource_api.get_project(domain['id'])
        with mock.patch.object(
            resource_sql.Resource, "get_project"
        ) as mock_get_project:
            mock_get_project.return_value = domain_ref
            # Delete the domain
            PROVIDERS.resource_api.delete_domain(domain['id'])
            self.assertEqual(mock_get_project.call_count, 1)

    @unit.skip_if_no_multiple_domains_support
    def test_domain_name_case_sensitivity(self):
        # create a ref with a lowercase name
        domain_name = 'test_domain'
        ref = unit.new_domain_ref(name=domain_name)

        lower_case_domain = PROVIDERS.resource_api.create_domain(
            ref['id'], ref
        )

        # assign a new ID to the ref with the same name, but in uppercase
        ref['id'] = uuid.uuid4().hex
        ref['name'] = domain_name.upper()
        upper_case_domain = PROVIDERS.resource_api.create_domain(
            ref['id'], ref
        )

        # We can get each domain by name
        lower_case_domain_ref = PROVIDERS.resource_api.get_domain_by_name(
            domain_name
        )
        self.assertDictEqual(lower_case_domain, lower_case_domain_ref)

        upper_case_domain_ref = PROVIDERS.resource_api.get_domain_by_name(
            domain_name.upper()
        )
        self.assertDictEqual(upper_case_domain, upper_case_domain_ref)

    def test_project_attribute_update(self):
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        PROVIDERS.resource_api.create_project(project['id'], project)

        # pick a key known to be non-existent
        key = 'description'

        def assert_key_equals(value):
            project_ref = PROVIDERS.resource_api.update_project(
                project['id'], project
            )
            self.assertEqual(value, project_ref[key])
            project_ref = PROVIDERS.resource_api.get_project(project['id'])
            self.assertEqual(value, project_ref[key])

        def assert_get_key_is(value):
            project_ref = PROVIDERS.resource_api.update_project(
                project['id'], project
            )
            self.assertIs(project_ref.get(key), value)
            project_ref = PROVIDERS.resource_api.get_project(project['id'])
            self.assertIs(project_ref.get(key), value)

        # add an attribute that doesn't exist, set it to a falsey value
        value = ''
        project[key] = value
        assert_key_equals(value)

        # set an attribute with a falsey value to null
        value = None
        project[key] = value
        assert_get_key_is(value)

        # do it again, in case updating from this situation is handled oddly
        value = None
        project[key] = value
        assert_get_key_is(value)

        # set a possibly-null value to a falsey value
        value = ''
        project[key] = value
        assert_key_equals(value)

        # set a falsey value to a truthy value
        value = uuid.uuid4().hex
        project[key] = value
        assert_key_equals(value)

    @unit.skip_if_cache_disabled('resource')
    @unit.skip_if_no_multiple_domains_support
    def test_domain_rename_invalidates_get_domain_by_name_cache(self):
        domain = unit.new_domain_ref()
        domain_id = domain['id']
        domain_name = domain['name']
        PROVIDERS.resource_api.create_domain(domain_id, domain)
        domain_ref = PROVIDERS.resource_api.get_domain_by_name(domain_name)
        domain_ref['name'] = uuid.uuid4().hex
        PROVIDERS.resource_api.update_domain(domain_id, domain_ref)
        self.assertRaises(
            exception.DomainNotFound,
            PROVIDERS.resource_api.get_domain_by_name,
            domain_name,
        )

    @unit.skip_if_cache_disabled('resource')
    def test_cache_layer_domain_crud(self):
        domain = unit.new_domain_ref()
        domain_id = domain['id']
        # Create Domain
        PROVIDERS.resource_api.create_domain(domain_id, domain)
        project_domain_ref = PROVIDERS.resource_api.get_project(domain_id)
        domain_ref = PROVIDERS.resource_api.get_domain(domain_id)
        updated_project_domain_ref = copy.deepcopy(project_domain_ref)
        updated_project_domain_ref['name'] = uuid.uuid4().hex
        updated_domain_ref = copy.deepcopy(domain_ref)
        updated_domain_ref['name'] = updated_project_domain_ref['name']
        # Update domain, bypassing resource api manager
        PROVIDERS.resource_api.driver.update_project(
            domain_id, updated_project_domain_ref
        )
        # Verify get_domain still returns the domain
        self.assertLessEqual(
            domain_ref.items(),
            PROVIDERS.resource_api.get_domain(domain_id).items(),
        )
        # Invalidate cache
        PROVIDERS.resource_api.get_domain.invalidate(
            PROVIDERS.resource_api, domain_id
        )
        # Verify get_domain returns the updated domain
        self.assertLessEqual(
            updated_domain_ref.items(),
            PROVIDERS.resource_api.get_domain(domain_id).items(),
        )
        # Update the domain back to original ref, using the assignment api
        # manager
        PROVIDERS.resource_api.update_domain(domain_id, domain_ref)
        self.assertLessEqual(
            domain_ref.items(),
            PROVIDERS.resource_api.get_domain(domain_id).items(),
        )
        # Make sure domain is 'disabled', bypass resource api manager
        project_domain_ref_disabled = project_domain_ref.copy()
        project_domain_ref_disabled['enabled'] = False
        PROVIDERS.resource_api.driver.update_project(
            domain_id, project_domain_ref_disabled
        )
        PROVIDERS.resource_api.driver.update_project(
            domain_id, {'enabled': False}
        )
        # Delete domain, bypassing resource api manager
        PROVIDERS.resource_api.driver.delete_project(domain_id)
        # Verify get_domain still returns the domain
        self.assertLessEqual(
            domain_ref.items(),
            PROVIDERS.resource_api.get_domain(domain_id).items(),
        )
        # Invalidate cache
        PROVIDERS.resource_api.get_domain.invalidate(
            PROVIDERS.resource_api, domain_id
        )
        # Verify get_domain now raises DomainNotFound
        self.assertRaises(
            exception.DomainNotFound,
            PROVIDERS.resource_api.get_domain,
            domain_id,
        )
        # Recreate Domain
        PROVIDERS.resource_api.create_domain(domain_id, domain)
        PROVIDERS.resource_api.get_domain(domain_id)
        # Make sure domain is 'disabled', bypass resource api manager
        domain['enabled'] = False
        PROVIDERS.resource_api.driver.update_project(domain_id, domain)
        PROVIDERS.resource_api.driver.update_project(
            domain_id, {'enabled': False}
        )
        # Delete domain
        PROVIDERS.resource_api.delete_domain(domain_id)
        # verify DomainNotFound raised
        self.assertRaises(
            exception.DomainNotFound,
            PROVIDERS.resource_api.get_domain,
            domain_id,
        )

    @unit.skip_if_cache_disabled('resource')
    @unit.skip_if_no_multiple_domains_support
    def test_project_rename_invalidates_get_project_by_name_cache(self):
        domain = unit.new_domain_ref()
        project = unit.new_project_ref(domain_id=domain['id'])
        project_id = project['id']
        project_name = project['name']
        PROVIDERS.resource_api.create_domain(domain['id'], domain)
        # Create a project
        PROVIDERS.resource_api.create_project(project_id, project)
        PROVIDERS.resource_api.get_project_by_name(project_name, domain['id'])
        project['name'] = uuid.uuid4().hex
        PROVIDERS.resource_api.update_project(project_id, project)
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.get_project_by_name,
            project_name,
            domain['id'],
        )

    @unit.skip_if_cache_disabled('resource')
    @unit.skip_if_no_multiple_domains_support
    def test_cache_layer_project_crud(self):
        domain = unit.new_domain_ref()
        project = unit.new_project_ref(domain_id=domain['id'])
        project_id = project['id']
        PROVIDERS.resource_api.create_domain(domain['id'], domain)
        # Create a project
        PROVIDERS.resource_api.create_project(project_id, project)
        PROVIDERS.resource_api.get_project(project_id)
        updated_project = copy.deepcopy(project)
        updated_project['name'] = uuid.uuid4().hex
        # Update project, bypassing resource manager
        PROVIDERS.resource_api.driver.update_project(
            project_id, updated_project
        )
        # Verify get_project still returns the original project_ref
        self.assertLessEqual(
            project.items(),
            PROVIDERS.resource_api.get_project(project_id).items(),
        )
        # Invalidate cache
        PROVIDERS.resource_api.get_project.invalidate(
            PROVIDERS.resource_api, project_id
        )
        # Verify get_project now returns the new project
        self.assertLessEqual(
            updated_project.items(),
            PROVIDERS.resource_api.get_project(project_id).items(),
        )
        # Update project using the resource_api manager back to original
        PROVIDERS.resource_api.update_project(project['id'], project)
        # Verify get_project returns the original project_ref
        self.assertLessEqual(
            project.items(),
            PROVIDERS.resource_api.get_project(project_id).items(),
        )
        # Delete project bypassing resource
        PROVIDERS.resource_api.driver.delete_project(project_id)
        # Verify get_project still returns the project_ref
        self.assertLessEqual(
            project.items(),
            PROVIDERS.resource_api.get_project(project_id).items(),
        )
        # Invalidate cache
        PROVIDERS.resource_api.get_project.invalidate(
            PROVIDERS.resource_api, project_id
        )
        # Verify ProjectNotFound now raised
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.get_project,
            project_id,
        )
        # recreate project
        PROVIDERS.resource_api.create_project(project_id, project)
        PROVIDERS.resource_api.get_project(project_id)
        # delete project
        PROVIDERS.resource_api.delete_project(project_id)
        # Verify ProjectNotFound is raised
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.get_project,
            project_id,
        )

    @unit.skip_if_no_multiple_domains_support
    def test_get_default_domain_by_name(self):
        domain_name = 'default'

        domain = unit.new_domain_ref(name=domain_name)
        PROVIDERS.resource_api.create_domain(domain['id'], domain)

        domain_ref = PROVIDERS.resource_api.get_domain_by_name(domain_name)
        self.assertEqual(domain, domain_ref)

    def test_get_not_default_domain_by_name(self):
        domain_name = 'foo'
        self.assertRaises(
            exception.DomainNotFound,
            PROVIDERS.resource_api.get_domain_by_name,
            domain_name,
        )

    def test_project_update_and_project_get_return_same_response(self):
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )

        PROVIDERS.resource_api.create_project(project['id'], project)

        updated_project = {'enabled': False}
        updated_project_ref = PROVIDERS.resource_api.update_project(
            project['id'], updated_project
        )

        # SQL backend adds 'extra' field
        updated_project_ref.pop('extra', None)

        self.assertIs(False, updated_project_ref['enabled'])

        project_ref = PROVIDERS.resource_api.get_project(project['id'])
        self.assertDictEqual(updated_project_ref, project_ref)

    def test_delete_project_clears_default_project_id(self):
        self.config_fixture.config(group='cache', enabled=False)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        user = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id, project_id=project['id']
        )
        PROVIDERS.resource_api.create_project(project['id'], project)
        user = PROVIDERS.identity_api.create_user(user)
        user = PROVIDERS.identity_api.get_user(user['id'])

        # LDAP is read only default_project_id doesn't exist
        if 'default_project_id' in user:
            self.assertIsNotNone(user['default_project_id'])
            PROVIDERS.resource_api.delete_project(project['id'])
            user = PROVIDERS.identity_api.get_user(user['id'])
            self.assertNotIn('default_project_id', user)

    def test_delete_project_with_roles_clears_default_project_id(self):
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        user = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id, project_id=project['id']
        )
        PROVIDERS.resource_api.create_project(project['id'], project)
        user = PROVIDERS.identity_api.create_user(user)
        role = unit.new_role_ref()
        PROVIDERS.role_api.create_role(role['id'], role)
        PROVIDERS.assignment_api.create_grant(
            user_id=user['id'], project_id=project['id'], role_id=role['id']
        )
        PROVIDERS.resource_api.delete_project(project['id'])
        user = PROVIDERS.identity_api.get_user(user['id'])
        self.assertNotIn('default_project_id', user)

    def _create_project_and_tags(self, num_of_tags=1):
        """Create a project and tags associated to that project.

        :param num_of_tags: the desired number of tags attached to a
                            project, default is 1.

        :returns: A tuple of a new project and a list of random tags
        """
        tags = [uuid.uuid4().hex for i in range(num_of_tags)]
        ref = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id, tags=tags
        )
        project = PROVIDERS.resource_api.create_project(ref['id'], ref)

        return project, tags

    def test_create_project_with_tags(self):
        project, tags = self._create_project_and_tags(num_of_tags=5)
        tag_ref = PROVIDERS.resource_api.get_project_tag(
            project['id'], tags[0]
        )
        self.assertEqual(tags[0], tag_ref)

    def test_get_project_contains_tags(self):
        project, _ = self._create_project_and_tags()
        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)
        ref = PROVIDERS.resource_api.get_project(project['id'])
        self.assertIn(tag, ref['tags'])

    def test_list_project_tags(self):
        project, tags = self._create_project_and_tags(num_of_tags=1)
        tag_ref = PROVIDERS.resource_api.list_project_tags(project['id'])
        self.assertEqual(tags[0], tag_ref[0])

    def test_list_project_tags_returns_not_found(self):
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.list_project_tags,
            uuid.uuid4().hex,
        )

    def test_get_project_tag(self):
        project, tags = self._create_project_and_tags()
        tag_ref = PROVIDERS.resource_api.get_project_tag(
            project['id'], tags[0]
        )
        self.assertEqual(tags[0], tag_ref)

    def test_create_project_tag_with_trailing_whitespace(self):
        project, _ = self._create_project_and_tags()
        tag = uuid.uuid4().hex + '   '
        tag_ref = PROVIDERS.resource_api.create_project_tag(project['id'], tag)
        self.assertEqual(tag.strip(), tag_ref)

    def test_create_project_tag_is_case_sensitive(self):
        project, _ = self._create_project_and_tags()
        new_tags = ['aaa', 'AAA']

        ref = PROVIDERS.resource_api.update_project_tags(
            project['id'], new_tags
        )
        for tag in new_tags:
            self.assertIn(tag, ref)

    def test_update_project_tags(self):
        project, tags = self._create_project_and_tags(num_of_tags=2)
        project_tag_ref = PROVIDERS.resource_api.list_project_tags(
            project['id']
        )
        self.assertEqual(len(project_tag_ref), 2)

        # Update project to only have one tag
        tags = ['one']
        PROVIDERS.resource_api.update_project_tags(project['id'], tags)
        project_tag_ref = PROVIDERS.resource_api.list_project_tags(
            project['id']
        )
        self.assertEqual(len(project_tag_ref), 1)

    def test_update_project_tags_returns_not_found(self):
        _, tags = self._create_project_and_tags(num_of_tags=2)
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.update_project_tags,
            uuid.uuid4().hex,
            tags,
        )

    def test_delete_tag_from_project(self):
        project, tags = self._create_project_and_tags(num_of_tags=2)
        tag_to_delete = tags[-1]
        PROVIDERS.resource_api.delete_project_tag(project['id'], tag_to_delete)
        project_tag_ref = PROVIDERS.resource_api.list_project_tags(
            project['id']
        )
        self.assertEqual(len(project_tag_ref), 1)
        self.assertEqual(project_tag_ref[0], tags[0])

    def test_delete_project_tag_returns_not_found(self):
        self.assertRaises(
            exception.ProjectNotFound,
            PROVIDERS.resource_api.delete_project_tag,
            uuid.uuid4().hex,
            uuid.uuid4().hex,
        )

    def test_delete_project_tags(self):
        project, tags = self._create_project_and_tags(num_of_tags=5)
        project_tag_ref = PROVIDERS.resource_api.list_project_tags(
            project['id']
        )
        self.assertEqual(len(project_tag_ref), 5)

        PROVIDERS.resource_api.update_project_tags(project['id'], [])
        project_tag_ref = PROVIDERS.resource_api.list_project_tags(
            project['id']
        )
        self.assertEqual(project_tag_ref, [])

    def test_create_project_immutable(self):
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project['options'][ro_opt.IMMUTABLE_OPT.option_name] = True

        p_created = PROVIDERS.resource_api.create_project(
            project['id'], project
        )
        project_via_manager = PROVIDERS.resource_api.get_project(project['id'])
        self.assertTrue('options' in p_created)
        self.assertTrue('options' in project_via_manager)
        self.assertTrue(
            project_via_manager['options'][ro_opt.IMMUTABLE_OPT.option_name]
        )
        self.assertTrue(p_created['options'][ro_opt.IMMUTABLE_OPT.option_name])

    def test_cannot_update_immutable_project(self):
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project['options'][ro_opt.IMMUTABLE_OPT.option_name] = True
        PROVIDERS.resource_api.create_project(project['id'], project)

        update_project = {'name': uuid.uuid4().hex}
        self.assertRaises(
            exception.ResourceUpdateForbidden,
            PROVIDERS.resource_api.update_project,
            project['id'],
            update_project,
        )

    def test_cannot_update_immutable_project_while_unsetting_immutable(self):
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project['options'][ro_opt.IMMUTABLE_OPT.option_name] = True
        PROVIDERS.resource_api.create_project(project['id'], project)

        update_project = {
            'name': uuid.uuid4().hex,
            'options': {ro_opt.IMMUTABLE_OPT.option_name: True},
        }
        self.assertRaises(
            exception.ResourceUpdateForbidden,
            PROVIDERS.resource_api.update_project,
            project['id'],
            update_project,
        )

    def test_cannot_delete_immutable_project(self):
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project['options'][ro_opt.IMMUTABLE_OPT.option_name] = True
        PROVIDERS.resource_api.create_project(project['id'], project)
        self.assertRaises(
            exception.ResourceDeleteForbidden,
            PROVIDERS.resource_api.delete_project,
            project['id'],
        )

    def test_update_project_set_immutable(self):
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        PROVIDERS.resource_api.create_project(project['id'], project)
        update_project = {'options': {ro_opt.IMMUTABLE_OPT.option_name: True}}
        project_via_manager = PROVIDERS.resource_api.get_project(project['id'])
        self.assertTrue('options' in project_via_manager)
        self.assertFalse(
            ro_opt.IMMUTABLE_OPT.option_name in project_via_manager['options']
        )
        p_update = PROVIDERS.resource_api.update_project(
            project['id'], update_project
        )
        project_via_manager = PROVIDERS.resource_api.get_project(project['id'])
        self.assertTrue(
            ro_opt.IMMUTABLE_OPT.option_name in p_update['options']
        )
        self.assertTrue(p_update['options'][ro_opt.IMMUTABLE_OPT.option_name])
        self.assertTrue(
            ro_opt.IMMUTABLE_OPT.option_name in project_via_manager['options']
        )
        self.assertTrue(
            project_via_manager['options'][ro_opt.IMMUTABLE_OPT.option_name]
        )

    def test_update_project_set_immutable_with_additional_updates(self):
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        PROVIDERS.resource_api.create_project(project['id'], project)
        update_project = {
            'name': uuid.uuid4().hex,
            'options': {ro_opt.IMMUTABLE_OPT.option_name: True},
        }
        project_via_manager = PROVIDERS.resource_api.get_project(project['id'])
        self.assertTrue('options' in project_via_manager)
        self.assertFalse(
            ro_opt.IMMUTABLE_OPT.option_name in project_via_manager['options']
        )
        p_update = PROVIDERS.resource_api.update_project(
            project['id'], update_project
        )
        project_via_manager = PROVIDERS.resource_api.get_project(project['id'])
        self.assertEqual(p_update['name'], update_project['name'])
        self.assertEqual(project_via_manager['name'], update_project['name'])
        self.assertTrue(
            ro_opt.IMMUTABLE_OPT.option_name in p_update['options']
        )
        self.assertTrue(p_update['options'][ro_opt.IMMUTABLE_OPT.option_name])
        self.assertTrue(
            ro_opt.IMMUTABLE_OPT.option_name in project_via_manager['options']
        )
        self.assertTrue(
            project_via_manager['options'][ro_opt.IMMUTABLE_OPT.option_name]
        )

    def test_update_project_unset_immutable(self):
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project['options'][ro_opt.IMMUTABLE_OPT.option_name] = True
        PROVIDERS.resource_api.create_project(project['id'], project)
        project_via_manager = PROVIDERS.resource_api.get_project(project['id'])
        self.assertTrue('options' in project_via_manager)
        self.assertTrue(
            project_via_manager['options'][ro_opt.IMMUTABLE_OPT.option_name]
        )

        update_project = {'options': {ro_opt.IMMUTABLE_OPT.option_name: False}}
        PROVIDERS.resource_api.update_project(project['id'], update_project)
        project_via_manager = PROVIDERS.resource_api.get_project(project['id'])
        self.assertTrue('options' in project_via_manager)
        self.assertTrue(
            ro_opt.IMMUTABLE_OPT.option_name in project_via_manager['options']
        )
        self.assertFalse(
            project_via_manager['options'][ro_opt.IMMUTABLE_OPT.option_name]
        )

        update_project = {'name': uuid.uuid4().hex}
        p_updated = PROVIDERS.resource_api.update_project(
            project['id'], update_project
        )
        self.assertEqual(p_updated['name'], update_project['name'])

        update_project = {'options': {ro_opt.IMMUTABLE_OPT.option_name: None}}
        p_updated = PROVIDERS.resource_api.update_project(
            project['id'], update_project
        )
        project_via_manager = PROVIDERS.resource_api.get_project(project['id'])
        self.assertTrue('options' in p_updated)
        self.assertTrue('options' in project_via_manager)
        self.assertFalse(
            ro_opt.IMMUTABLE_OPT.option_name in p_updated['options']
        )
        self.assertFalse(
            ro_opt.IMMUTABLE_OPT.option_name in project_via_manager['options']
        )

    def test_cannot_delete_project_tags_immutable_project(self):
        project, tags = self._create_project_and_tags(num_of_tags=2)
        update_project = {'options': {ro_opt.IMMUTABLE_OPT.option_name: True}}
        PROVIDERS.resource_api.update_project(project['id'], update_project)
        self.assertRaises(
            exception.ResourceUpdateForbidden,
            PROVIDERS.resource_api.delete_project_tag,
            project['id'],
            tags[0],
        )

    def test_cannot_update_project_tags_immutable_project(self):
        # Update and Add tag use the same API
        project, tags = self._create_project_and_tags(num_of_tags=2)
        update_project = {'options': {ro_opt.IMMUTABLE_OPT.option_name: True}}
        PROVIDERS.resource_api.update_project(project['id'], update_project)
        tags.append(uuid.uuid4().hex)
        self.assertRaises(
            exception.ResourceUpdateForbidden,
            PROVIDERS.resource_api.update_project_tags,
            project['id'],
            tags,
        )

    @unit.skip_if_no_multiple_domains_support
    def test_create_domain_immutable(self):
        domain_id = uuid.uuid4().hex

        domain = {
            'name': uuid.uuid4().hex,
            'id': domain_id,
            'is_domain': True,
            'options': {'immutable': True},
        }

        PROVIDERS.resource_api.create_domain(domain_id, domain)
        domain_via_manager = PROVIDERS.resource_api.get_domain(domain_id)
        self.assertTrue('options' in domain_via_manager)
        self.assertTrue(domain_via_manager['options']['immutable'])

    @unit.skip_if_no_multiple_domains_support
    def test_cannot_update_immutable_domain(self):
        domain_id = uuid.uuid4().hex

        domain = {
            'name': uuid.uuid4().hex,
            'id': domain_id,
            'is_domain': True,
            'options': {'immutable': True},
        }

        PROVIDERS.resource_api.create_domain(domain_id, domain)
        update_domain = {'name': uuid.uuid4().hex}
        self.assertRaises(
            exception.ResourceUpdateForbidden,
            PROVIDERS.resource_api.update_domain,
            domain_id,
            update_domain,
        )

    @unit.skip_if_no_multiple_domains_support
    def test_cannot_delete_immutable_domain(self):
        domain_id = uuid.uuid4().hex

        domain = {
            'name': uuid.uuid4().hex,
            'id': domain_id,
            'is_domain': True,
            'options': {'immutable': True},
        }

        PROVIDERS.resource_api.create_domain(domain_id, domain)
        self.assertRaises(
            exception.ResourceDeleteForbidden,
            PROVIDERS.resource_api.delete_domain,
            domain_id,
        )

    @unit.skip_if_no_multiple_domains_support
    def test_cannot_delete_disabled_domain_with_immutable_project(self):
        domain_id = uuid.uuid4().hex

        domain = {'name': uuid.uuid4().hex, 'id': domain_id, 'is_domain': True}

        PROVIDERS.resource_api.create_domain(domain_id, domain)
        project = unit.new_project_ref(domain_id)
        project['options'][ro_opt.IMMUTABLE_OPT.option_name] = True
        PROVIDERS.resource_api.create_project(project['id'], project)
        # Disable the domain
        PROVIDERS.resource_api.update_domain(domain_id, {'enabled': False})
        # attempt to delete the domain, should error when the immutable
        # project is reached
        self.assertRaises(
            exception.ResourceDeleteForbidden,
            PROVIDERS.resource_api.delete_domain,
            domain_id,
        )

    @unit.skip_if_no_multiple_domains_support
    def test_update_domain_set_immutable(self):
        # domains are projects, this should be the same as the project version
        domain_id = uuid.uuid4().hex

        domain = {'name': uuid.uuid4().hex, 'id': domain_id, 'is_domain': True}

        PROVIDERS.resource_api.create_domain(domain_id, domain)
        domain_via_manager = PROVIDERS.resource_api.get_domain(domain_id)
        self.assertTrue('options' in domain_via_manager)
        self.assertFalse(
            ro_opt.IMMUTABLE_OPT.option_name in domain_via_manager['options']
        )

        domain_update = {'options': {ro_opt.IMMUTABLE_OPT.option_name: True}}
        d_update = PROVIDERS.resource_api.update_domain(
            domain_id, domain_update
        )
        domain_via_manager = PROVIDERS.resource_api.get_domain(domain_id)
        self.assertTrue(
            ro_opt.IMMUTABLE_OPT.option_name in d_update['options']
        )
        self.assertTrue(d_update['options'][ro_opt.IMMUTABLE_OPT.option_name])
        self.assertTrue(
            ro_opt.IMMUTABLE_OPT.option_name in domain_via_manager['options']
        )
        self.assertTrue(
            domain_via_manager['options'][ro_opt.IMMUTABLE_OPT.option_name]
        )

    def test_update_domain_unset_immutable(self):
        # domains are projects, this should be the same as the project version
        domain_id = uuid.uuid4().hex

        domain = {
            'name': uuid.uuid4().hex,
            'id': domain_id,
            'is_domain': True,
            'options': {ro_opt.IMMUTABLE_OPT.option_name: True},
        }

        PROVIDERS.resource_api.create_domain(domain_id, domain)
        domain_via_manager = PROVIDERS.resource_api.get_domain(domain_id)
        self.assertTrue('options' in domain_via_manager)
        self.assertTrue(
            domain_via_manager['options'][ro_opt.IMMUTABLE_OPT.option_name]
        )

        update_domain = {'options': {ro_opt.IMMUTABLE_OPT.option_name: False}}
        d_updated = PROVIDERS.resource_api.update_domain(
            domain_id, update_domain
        )
        domain_via_manager = PROVIDERS.resource_api.get_domain(domain_id)
        self.assertTrue('options' in domain_via_manager)
        self.assertTrue('options' in d_updated)
        self.assertTrue(
            ro_opt.IMMUTABLE_OPT.option_name in domain_via_manager['options']
        )
        self.assertTrue(
            ro_opt.IMMUTABLE_OPT.option_name in d_updated['options']
        )
        self.assertFalse(
            d_updated['options'][ro_opt.IMMUTABLE_OPT.option_name]
        )
        self.assertFalse(
            domain_via_manager['options'][ro_opt.IMMUTABLE_OPT.option_name]
        )

        update_domain = {'name': uuid.uuid4().hex}
        d_updated = PROVIDERS.resource_api.update_domain(
            domain_id, update_domain
        )
        self.assertEqual(d_updated['name'], update_domain['name'])

        update_domain = {'options': {ro_opt.IMMUTABLE_OPT.option_name: None}}
        d_updated = PROVIDERS.resource_api.update_domain(
            domain_id, update_domain
        )
        domain_via_manager = PROVIDERS.resource_api.get_domain(domain_id)
        self.assertTrue('options' in d_updated)
        self.assertTrue('options' in domain_via_manager)
        self.assertFalse(
            ro_opt.IMMUTABLE_OPT.option_name in d_updated['options']
        )
        self.assertFalse(
            ro_opt.IMMUTABLE_OPT.option_name in domain_via_manager['options']
        )


class ResourceDriverTests:
    """Test for the resource driver.

    Subclasses must set self.driver to the driver instance.

    """

    def test_create_project(self):
        project_id = uuid.uuid4().hex
        project = {
            'name': uuid.uuid4().hex,
            'id': project_id,
            'domain_id': default_fixtures.ROOT_DOMAIN['id'],
        }
        self.driver.create_project(project_id, project)

    def test_create_project_all_defined_properties(self):
        project_id = uuid.uuid4().hex
        project = {
            'name': uuid.uuid4().hex,
            'id': project_id,
            'domain_id': default_fixtures.ROOT_DOMAIN['id'],
        }
        parent_project = self.driver.create_project(project_id, project)

        project_id = uuid.uuid4().hex
        project = {
            'name': uuid.uuid4().hex,
            'id': project_id,
            'domain_id': default_fixtures.ROOT_DOMAIN['id'],
            'description': uuid.uuid4().hex,
            'enabled': True,
            'parent_id': parent_project['id'],
            'is_domain': True,
        }
        self.driver.create_project(project_id, project)

    def test_create_project_null_domain(self):
        project_id = uuid.uuid4().hex
        project = {
            'name': uuid.uuid4().hex,
            'id': project_id,
            'domain_id': None,
        }
        self.driver.create_project(project_id, project)

    def test_create_project_same_name_same_domain_conflict(self):
        name = uuid.uuid4().hex
        domain_id = default_fixtures.ROOT_DOMAIN['id']

        project_id = uuid.uuid4().hex
        project = {'name': name, 'id': project_id, 'domain_id': domain_id}
        self.driver.create_project(project_id, project)

        project_id = uuid.uuid4().hex
        project = {'name': name, 'id': project_id, 'domain_id': domain_id}
        self.assertRaises(
            exception.Conflict, self.driver.create_project, project_id, project
        )

    def test_create_project_same_id_conflict(self):
        project_id = uuid.uuid4().hex

        project = {
            'name': uuid.uuid4().hex,
            'id': project_id,
            'domain_id': default_fixtures.ROOT_DOMAIN['id'],
        }
        self.driver.create_project(project_id, project)

        project = {
            'name': uuid.uuid4().hex,
            'id': project_id,
            'domain_id': default_fixtures.ROOT_DOMAIN['id'],
        }
        self.assertRaises(
            exception.Conflict, self.driver.create_project, project_id, project
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from keystone.conf import utils

driver = cfg.StrOpt(
    'driver',
    default='sql',
    help=utils.fmt(
        """
Entry point for the identity mapping backend driver in the
`keystone.identity.id_mapping` namespace. Keystone only provides a `sql`
driver, so there is no reason to change this unless you are providing a custom
entry point.
"""
    ),
)

generator = cfg.StrOpt(
    'generator',
    default='sha256',
    help=utils.fmt(
        """
Entry point for the public ID generator for user and group entities in the
`keystone.identity.id_generator` namespace. The Keystone identity mapper only
supports generators that produce 64 bytes or less. Keystone only provides a
`sha256` entry point, so there is no reason to change this value unless you're
providing a custom entry point.
"""
    ),
)

backward_compatible_ids = cfg.BoolOpt(
    'backward_compatible_ids',
    default=True,
    help=utils.fmt(
        """
The format of user and group IDs changed in Juno for backends that do not
generate UUIDs (for example, LDAP), with keystone providing a hash mapping to
the underlying attribute in LDAP. By default this mapping is disabled, which
ensures that existing IDs will not change. Even when the mapping is enabled by
using domain-specific drivers (`[identity] domain_specific_drivers_enabled`),
any users and groups from the default domain being handled by LDAP will still
not be mapped to ensure their IDs remain backward compatible. Setting this
value to false will enable the new mapping for all backends, including the
default LDAP driver. It is only guaranteed to be safe to enable this option
if you do not already have assignments for users and groups from the default
LDAP domain, and you consider it to be acceptable for Keystone to provide the
different IDs to clients than it did previously (existing IDs in the API will
suddenly change). Typically this means that the only time you can set this
value to false is when configuring a fresh installation, although that is the
recommended value.
"""
    ),
)


GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [driver, generator, backward_compatible_ids]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from sqlalchemy.ext import declarative

from keystone.common import sql
from keystone.tests import unit
from keystone.tests.unit import utils

ModelBase = declarative.declarative_base()


class TestModel(ModelBase, sql.ModelDictMixin):  # type: ignore
    __tablename__ = 'testmodel'
    id = sql.Column(sql.String(64), primary_key=True)
    text = sql.Column(sql.String(64), nullable=False)


class TestModelDictMixin(unit.BaseTestCase):
    def test_creating_a_model_instance_from_a_dict(self):
        d = {'id': utils.new_uuid(), 'text': utils.new_uuid()}
        m = TestModel.from_dict(d)
        self.assertEqual(d['id'], m.id)
        self.assertEqual(d['text'], m.text)

    def test_creating_a_dict_from_a_model_instance(self):
        m = TestModel(id=utils.new_uuid(), text=utils.new_uuid())
        d = m.to_dict()
        self.assertEqual(d['id'], m.id)
        self.assertEqual(d['text'], m.text)

    def test_creating_a_model_instance_from_an_invalid_dict(self):
        d = {'id': utils.new_uuid(), 'text': utils.new_uuid(), 'extra': None}
        self.assertRaises(TypeError, TestModel.from_dict, d)

    def test_creating_a_dict_from_a_model_instance_that_has_extra_attrs(self):
        expected = {'id': utils.new_uuid(), 'text': utils.new_uuid()}
        m = TestModel(id=expected['id'], text=expected['text'])
        m.extra = 'this should not be in the dictionary'
        # NOTE(notmorgan): This is currently explicitly harmless as this does
        # not actually use SQL-Alchemy.
        self.assertEqual(expected, m.to_dict())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_policy import policy

from keystone.common.policies import base

registered_limit_policies = [
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_registered_limit',
        check_str='',
        scope_types=['system', 'domain', 'project'],
        description='Show registered limit details.',
        operations=[
            {
                'path': '/v3/registered_limits/{registered_limit_id}',
                'method': 'GET',
            },
            {
                'path': '/v3/registered_limits/{registered_limit_id}',
                'method': 'HEAD',
            },
        ],
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_registered_limits',
        check_str='',
        scope_types=['system', 'domain', 'project'],
        description='List registered limits.',
        operations=[
            {'path': '/v3/registered_limits', 'method': 'GET'},
            {'path': '/v3/registered_limits', 'method': 'HEAD'},
        ],
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'create_registered_limits',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Create registered limits.',
        operations=[{'path': '/v3/registered_limits', 'method': 'POST'}],
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'update_registered_limit',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Update registered limit.',
        operations=[
            {
                'path': '/v3/registered_limits/{registered_limit_id}',
                'method': 'PATCH',
            }
        ],
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'delete_registered_limit',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Delete registered limit.',
        operations=[
            {
                'path': '/v3/registered_limits/{registered_limit_id}',
                'method': 'DELETE',
            }
        ],
    ),
]


def list_rules():
    return registered_limit_policies



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Initial no-op Yoga contract migration.

Revision ID: e25ffa003242
Revises: 27e647c0fad4
Create Date: 2022-01-21 00:00:00.000000
"""

# revision identifiers, used by Alembic.
revision = 'e25ffa003242'
down_revision = '27e647c0fad4'
branch_labels = ('contract',)
# milestone identifier
keystone_milestone = ['yoga']


def upgrade():
    pass



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import random
import secrets
import string

from oslo_config import fixture as config_fixture

from keystone.common import password_hashing
import keystone.conf
from keystone.tests import unit

CONF = keystone.conf.CONF


class TestPasswordHashing(unit.BaseTestCase):
    OPTIONAL = object()
    ITERATIONS: int = 10

    def setUp(self):
        super().setUp()
        self.config_fixture = self.useFixture(config_fixture.Config(CONF))

    def test_scrypt(self):
        self.config_fixture.config(strict_password_check=True)
        self.config_fixture.config(
            group="identity", password_hash_algorithm="scrypt"
        )
        self.config_fixture.config(group="identity", max_password_length="96")
        # Few iterations to test different inputs
        for _ in range(self.ITERATIONS):
            password: str = "".join(  # type: ignore
                secrets.choice(string.printable)
                for i in range(random.randint(1, 72))
            )
            hashed = password_hashing.hash_password(password)
            self.assertTrue(password_hashing.check_password(password, hashed))

    def test_bcrypt(self):
        self.config_fixture.config(strict_password_check=True)
        self.config_fixture.config(
            group="identity", password_hash_algorithm="bcrypt"
        )
        self.config_fixture.config(group="identity", max_password_length="72")
        # Few iterations to test different inputs
        for _ in range(self.ITERATIONS):
            password: str = "".join(  # type: ignore
                secrets.choice(string.printable)
                for i in range(random.randint(1, 72))
            )
            hashed = password_hashing.hash_password(password)
            self.assertTrue(password_hashing.check_password(password, hashed))

    def test_bcrypt_sha256(self):
        self.config_fixture.config(strict_password_check=True)
        self.config_fixture.config(
            group="identity", password_hash_algorithm="bcrypt_sha256"
        )
        self.config_fixture.config(group="identity", max_password_length="96")
        # Few iterations to test different inputs
        for _ in range(self.ITERATIONS):
            password: str = "".join(  # type: ignore
                secrets.choice(string.printable)
                for i in range(random.randint(1, 96))
            )
            hashed = password_hashing.hash_password(password)
            self.assertTrue(password_hashing.check_password(password, hashed))

    def test_pbkdf2_sha512(self):
        self.config_fixture.config(strict_password_check=True)
        self.config_fixture.config(
            group="identity", password_hash_algorithm="pbkdf2_sha512"
        )
        self.config_fixture.config(group="identity", max_password_length="96")
        # Do few iterations to process different inputs
        for _ in range(self.ITERATIONS):
            password: str = "".join(  # type: ignore
                secrets.choice(string.printable)
                for i in range(random.randint(1, 96))
            )
            hashed = password_hashing.hash_password(password)
            self.assertTrue(password_hashing.check_password(password, hashed))
